/*
 * check_direct () - after site failure, check the directories to determine
 *		if more copies of an object need to be created.
 */
#include <stdio.h>
#include "comm.h"
#include "simtypes.h"
#include "extdata.h"
#include "direct.h"

/* these will hold the new directories and fail locks (if necessary) */
directory_type	new_directories;
fail_locks_type	new_fail_locks;

int check_direct ()
{
int	directory_changed = FALSE;
int	site_failed = FALSE;		/* just to do it the first time */
int	num_good_objects;
site_id_type	site;
data_item_type	item;
site_state_type get_site_state();
msg_type *dir_msg, *fl_msg;		/* bogus messages for processing */

do {		/* try to find out if new objects need to be created */
    directory_changed = site_failed = FALSE;

    printf("Checking directories \n" ) ; 
    /* start with the old directories and fail locks */
    for (item = 0; item < MAX_NUM_ITEMS; item++) {
	new_directories[item].direct = directories[item].direct;
	new_directories[item].num_copies = directories[item].num_copies;
	new_directories[item].min_copies = directories[item].min_copies;
	new_fail_locks[item] = fail_locks[item];
    }

    /* look at each item in the database to determine if there are */
    /* enough GOOD copies (no fail lock) on up sites for the required */
    /* minimum specified */

    for (item = 0; item < num_db_items; item++ ) {
	for (num_good_objects = 0, site = FIRST_SITE_ID; 
		site < num_sites; site++) {

		if (get_site_state(site) == SITE_UP
		 && directory_mgr (CHECK_FOR_COPY, site, item) == COPY_PRESENT 
		 && fail_locks_mgr (CHECK_FOR_SITE, site, item) == FAIL_LOCK_CLEAR) {
			num_good_objects++;
		}
	}
	/* did we end up with enough copies ??? and is there somewhere 
	 * to get new copies from?
	 */
	while (num_good_objects != 0 && 
	       num_good_objects < directories[item].min_copies) {

	    if (create_new_object(item) == TRUE) {
		num_good_objects++;
		directory_changed = TRUE;
	    }else
		num_good_objects = 0;  /* no more sites available for creation*/
	}

    }

    /* new directories and fail locks table have been created */
    /* or have they? (nothing may have changed) */

    if (directory_changed)
	site_failed = install_changes();

}while (directory_changed && site_failed);

if (directory_changed) {
    printf("Directory Changed\n") ; 
    dir_msg = (msg_type *) malloc (sizeof  (msg_type));
    fl_msg = (msg_type *) malloc (sizeof (msg_type));
    bcopy (new_directories, dir_msg->c.ctl_direct, sizeof (directory_type));
    bcopy (new_fail_locks, fl_msg->c.ctl.f_locks, sizeof (fail_locks_type));

    /* pass along the bogus messages to the routine to move the new */
    /* directories and fail locks into place.   */
    get_new_copies (dir_msg, fl_msg);
    log_site_info(DUMP, "dumping after directory update") ; 
}

return(0);
}

/*
 * install_changes() - run a two phase commit protocol to commit the new
 *			directories and fail lock tables
 */
int install_changes ()
{
int send_message(), receive_message(), announce_site_failure();
site_state_type set_site_state(), get_site_state();
site_id_type  site;
msg_type *upd_msg, *snd_msg, *rcv_msg, *abort_msg;
int commit_transaction = TRUE,  failure_detected;

    upd_msg = (msg_type *) malloc (sizeof(msg_type));
    upd_msg->source = MySID;
    upd_msg->type = xact;
    upd_msg->subtype = direct_update;

    rcv_msg = (msg_type *) malloc (sizeof(msg_type));

    for (site = FIRST_SITE_ID; site < num_sites; site++) {
	if ((site != MySID) && (get_site_state(site) == SITE_UP)) {
	    upd_msg->dest = site;
	    if (send_message(upd_msg) != ERROR) {
		receive_message (rcv_msg);

		if ((rcv_msg->type == managing) &&
		    (rcv_msg->subtype == failed)) {
		    commit_transaction = FALSE;
		    failure_detected = TRUE;
		    set_site_state (rcv_msg->source, SITE_DOWN);
		    announce_site_failure();
		    break;
		}else if ((rcv_msg->type != xact) || (rcv_msg->subtype != ack))
			log_site_info (NO_DUMP, "*** unexpected message ***");
	    }
        }
    }    

    /* round one completed - if directory xact has not been aborted then
     * start round two of the commit
     */
    if (commit_transaction) {
	snd_msg = (msg_type *) malloc (sizeof(msg_type));
	snd_msg->source = MySID;
	snd_msg->type = xact;
	snd_msg->subtype = commit;

	failure_detected = FALSE;
	for (site = FIRST_SITE_ID; site < num_sites; site++) {
	    if ((site != MySID) && (get_site_state(site) == SITE_UP)){
		snd_msg->dest = site;
		if (send_message (snd_msg) != ERROR) {
		    receive_message (rcv_msg);
		    if ((rcv_msg->type == managing) &&
			(rcv_msg->subtype == failed)) {
			failure_detected = TRUE;
			set_site_state (rcv_msg->source, SITE_DOWN);
		    }else
		       if ((rcv_msg->type != xact) ||
			   (rcv_msg->subtype != commit_ack)) {
			   log_site_info (NO_DUMP, "*** unexpected message ***");
		    }
		}
	    }
	}
	free (rcv_msg);
	if (failure_detected) announce_site_failure ();
    }
    free (upd_msg);
    if (!failure_detected) {
	snd_msg->source = MySID;
	snd_msg->type = control;
	snd_msg->subtype = recovery_response;
	bcopy(session_vector, snd_msg->c.ctl.s_vector, sizeof(session_vector_type));
	bcopy(new_fail_locks, snd_msg->c.ctl.f_locks, sizeof(fail_locks_type));
	for (site = FIRST_SITE_ID; site < num_sites; site++) {
	    if ((site != MySID) && (get_site_state(site) == SITE_UP)){ 
		snd_msg->dest = site ; 
		send_message(snd_msg);
	    } 
	}
	/* send the new fail-locks to the manager too */

	snd_msg->dest = SIMULATION_MGR_ID ; 
	send_message(snd_msg) ; 

	snd_msg->source = MySID;
	snd_msg->type = control;
	snd_msg->subtype = directory_ack;
	bcopy(new_directories, snd_msg->c.ctl_direct, sizeof(directory_type));
	for (site = FIRST_SITE_ID; site < num_sites; site++) {
	    if ((site != MySID) && (get_site_state(site) == SITE_UP)) { 
		snd_msg->dest = site ; 
		send_message(snd_msg);
	    } 
	}
	/* send the new directories to the manager too */

	snd_msg->dest = SIMULATION_MGR_ID ; 
	send_message(snd_msg) ; 

	free(snd_msg);
    }
    return (failure_detected);
}

/*
 * create_new_object () - attempt to create a new object on an up site
 */
int create_new_object (item)
int item ; 
{
int	object_created;
site_id_type	site, tmpsite;
site_state_type get_site_state();

    object_created = FALSE;		/* nothing created, yet */

    for (site = FIRST_SITE_ID; !object_created && site < num_sites; site++){

	if (get_site_state(site) == SITE_UP) {

	  /* is the fail lock cleared for this item? */
	  if (TEST_BIT(new_fail_locks[item], site)) {
		CLEAR_BIT(new_fail_locks[item], site);
		object_created = TRUE;
	  }else {  /* is the object already on this site? */
	      if (!TEST_BIT(new_directories[item].direct,site)) {

	   	SET_BIT(new_directories[item].direct, site);

		/* clear an object from ONE down site */
		for (tmpsite = FIRST_SITE_ID; tmpsite < num_sites; tmpsite++) {
		    if ((get_site_state(tmpsite) == SITE_DOWN) &&
		       (TEST_BIT(new_directories[item].direct, tmpsite))) {
		       CLEAR_BIT(new_directories[item].direct, tmpsite);
		       CLEAR_BIT(new_fail_locks[item], tmpsite);
		       break;		/* stop removing copies */
		    }
		}
		      
		object_created = TRUE;
	      }
	   }
	}
    }
    /* did we create a new object? */
    return (object_created);
}

/*
 * process_direct_update () - a xact , direct_update has been received
 *		this function will handle the rest of the two phase
 *		commit protocol.
 */
int process_direct_update (direct_updt, fail_locks, directs)
msg_type *direct_updt, **fail_locks, **directs;
{
int send_message(), receive_message();
site_state_type set_site_state(), get_site_state();
msg_type *cur_msg;
int need_final_reply = TRUE;

    cur_msg = (msg_type *) malloc (sizeof(msg_type));
    cur_msg->source = MySID;
    cur_msg->dest = direct_updt->source;
    cur_msg->type = xact;
    cur_msg->subtype = ack;
    /* respond to xact.direct_update */
    send_message (cur_msg);

    /* find out what coordinator wants  us to do */
    receive_message (cur_msg);
     
    switch (cur_msg->type) {
	case xact:
	    if (cur_msg->subtype != commit) {
		need_final_reply = FALSE;
		log_site_info (NO_DUMP, "*** unexepected message ***");
	    /* we don't have to fail so reply normally */
	    }else if (failure_schedule != FAIL_AFTER_UPD_RECV) {
		cur_msg->type = xact;
		cur_msg->subtype = commit_ack;
	    }else {  /* make things confusing and send a failure */
		set_site_state(MySID, SITE_DOWN);
		log_site_info (DUMP, "failing after update received");
		cur_msg->type = managing;
		cur_msg->subtype = failed;
	    }
	    break;
	case control:
	    /* somebody else has died, mark in our memory who it was */
	    need_final_reply = FALSE;
	    if (cur_msg->subtype == failure_announce)
		process_failure_announce (cur_msg);
	    else
		log_site_info (NO_DUMP, "*** unexpected message ***");
	    break;
	default:
	    need_final_reply = FALSE;
	    log_site_info (NO_DUMP, "*** unexpected message ***");
	    break;
    }

    if (need_final_reply) {
	cur_msg->source = MySID;
	cur_msg->dest = direct_updt->source;
	send_message (cur_msg);
	/* if we are still up... */
	if (get_site_state(MySID) == SITE_UP) {
	    *fail_locks = (msg_type *)malloc (sizeof (msg_type));
	    *directs = (msg_type *)malloc (sizeof (msg_type));

	    /* get the new fail locks and directories from coordinator */
	    receive_message(*fail_locks);
	    while ((*fail_locks)->type != control || (*fail_locks)->subtype
			!= recovery_response ) { 
		 if ( (*fail_locks)->type == xact && (*fail_locks)->subtype
			== copier ) 
		    process_copier_xact(*fail_locks) ; 
		 else 
		    printf ("ERROR in checkdirect\n") ; 
		 receive_message(*fail_locks) ;
	    } 
	    receive_message(*directs);
	    while ((*directs)->type != control || (*directs)->subtype
			!= directory_ack) { 
		 if ( (*directs)->type == xact && (*directs)->subtype
			== copier ) 
		    process_copier_xact(*directs) ; 
		 else 
		    printf ("ERROR in checkdirect\n") ; 
		 receive_message(*directs) ;
	    } 
	    return (TRUE);
	}
    }
    return (FALSE);
}
