/*---------------------------------------------------------------------------
 * comm.c -- cs542 communications
 * 
 * This package implements a library of calls which allow
 * network sites to be simulated and provides a facility whereby
 * sites can send messages (strings) to each other.
 *
 * When a site comes up it picks a UDP port on the local host and
 * writes it into a well-known file. Other sites can then read this
 * file to discover which port to send messages to for this site to
 * receive it.
 *
 * Cannibalized from raid communications library.
 *---------------------------------------------------------------------------
 */

/*---------------------------------------------------------------------------
 * Setup -- initialize communications
 * 
 * This routine finds a socket for the server to listen at, and 
 * advertises the port number in a well-known file.
 *
 *---------------------------------------------------------------------------
 */

/**/
#define O_CREAT 0x100			

#define O_RDWR  2

#define LOCK_EX 2
#define LOCK_UN 8
/**/					/* #defines added by DM */


#include "sys_comm.h"

int sMySock;				/* socket we listen for messages on */
struct sockaddr_in sa;			/* general purpose sa structure */
int AddrCache[MAX_SITES];		/* port number of each site */
int MySID;				/* name of this site */
char PortFile[BUFSIZ];			/* name of port file */
int PortFileFD;				/* fd of opened port file */
static time_t PFLastMtime = 0;		/* file last modify time */

/*
 * int alrm();
 * signal handlers should return void - DM
 */
void alrm();				

Setup(wSID)
int wSID;
{
	struct hostent *phe;
	int lsa;
	int iPort;
	char sbMyHost[BUFSIZ];
	
	
	signal(SIGALRM, alrm);
	if (wSID >= MAX_SITES) {
		fprintf(stderr, "site id must be in range [0,%d]\n", MAX_SITES);
		exit(ERROR);
	}
	MySID = wSID;

	/* get a socket */
	if ((sMySock = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		perror("socket");
		return(ERROR);
	}
	zero(sa);
	sa.sin_family = AF_INET;
	sa.sin_port = 0;
	/* find our host address */
	if (gethostname(sbMyHost, BUFSIZ) < 0) {
		perror("gethostname");
		return(ERROR);
	}
	if ((phe = gethostbyname(sbMyHost)) == NULL) {
		perror("gethostbyname");
		return(ERROR);
	}
	bcopy(phe->h_addr, &sa.sin_addr, phe->h_length);
	/* find a port to listen at */
	for (iPort = START_PORT; iPort < END_PORT; iPort++) {
		sa.sin_port = iPort;

/* Added cast to 'struct sockaddr *' for argument 2 of bind() - DM */
		if (bind(sMySock, (struct sockaddr *)&sa, sizeof(sa)) == 0)
			break;
	}
	if (iPort == END_PORT) {
		fprintf(stderr, "all ports busy. (!)\n");
		return(ERROR);
	}
	if (GetPortFileName() == ERROR)
		return(ERROR);
	InitAddrCache();
	if (OpenPortFile() == ERROR)
		return(ERROR);
	return(AdvertisePort(iPort));
}

/* 'void' return type added - DM */
void alrm()
{

}

/*---------------------------------------------------------------------------
 * InitAddrCache -- initialize the address cache
 *
 *---------------------------------------------------------------------------
 */

InitAddrCache()
{
	int i;

	for (i = 0; i < MAX_SITES; i++)
		AddrCache[i] = ERROR;	/* nonsense port number */
}

/*---------------------------------------------------------------------------
 * AdvertisePort -- place our port number in a well-known file
 *
 *---------------------------------------------------------------------------
 */

AdvertisePort(wPort)
int wPort;
{

	if (LockPortFile() == ERROR) 
		return(ERROR);
	ReadAddrCache();
	AddrCache[MySID] = wPort;
	WriteAddrCache();
	if (UnlockPortFile() == ERROR)
		return(ERROR);
	return(OK);
}

/*---------------------------------------------------------------------
 * CreatePortFile, DeletePortFile
 *-------------------------------------------------------------------*/

CreatePortFile()
{
     int  tmp_fd;

     GetPortFileName();
     if ((tmp_fd = open(PortFile, O_CREAT, 0666)) < 0) {
	  fprintf(stderr, "problems creating port file.\n");
	  return(ERROR);
     }
     close(tmp_fd);
     return(OK);
}

DeletePortFile()
{
     if (unlink(PortFile) < 0) {
	  fprintf(stderr, "problems deleting port file.\n");
	  return(ERROR);
     }
     return(OK);
}

/*---------------------------------------------------------------------------
 * OpenPortFile -- open the port file for reading and writing
 *
 * We obtain both a stream and a file descriptor, because certain
 * necessary operations on files are provided for one form and not
 * the other (setbuf .. ruddy UNIX .. ).
 * 
 *---------------------------------------------------------------------------
 */

OpenPortFile()
{
	FILE *fp;

	if ((PortFileFD = open(PortFile, O_RDWR, 0666)) < 0) {
		perror("open");
		fprintf(stderr, "problems opening port file.\n");
		return(ERROR);
	}
	if ((fp = fdopen(PortFileFD, "r+")) == NULL) {
		perror("fdopen");
		fprintf(stderr, "contact avr.\n");
		return(ERROR);
	}
	setbuf(fp, NULL);
	return(OK);
}

/*---------------------------------------------------------------------------
 * GetPortFileName -- return port file name
 *
 *---------------------------------------------------------------------------
 */

GetPortFileName()
{
	char *home;

/*---------------------------------------------------------------------
 *      Previously, ports file was in HOME directory.  Use default instead.
 *
 *	if ((home = (char *) getenv("HOME")) == NULL) {
 *		fprintf(stderr, "HOME environment variable not set.\n");
 *		return(ERROR);
 *	}
 *	sprintf(PortFile, "%s/ports", home);
 *-------------------------------------------------------------------*/

	sprintf(PortFile, "ports.tmp");
	return(OK);
}

/*---------------------------------------------------------------------------
 * LockPortFile -- lock port file with flock
 *
 *---------------------------------------------------------------------------
 */

LockPortFile()
{
/* need to change flock calls to fcntl calls somehow - DM */
/* http://www.ecst.csuchico.edu/~beej/guide/ipc/flock.html */

	/*if (flock(PortFileFD, LOCK_EX) == ERROR) { 
		perror("flock");
		fprintf(stderr, "attempt to lock port file failed.\n");
		fprintf(stderr, "contact avr.\n");
		return(ERROR);
	} */
 	return(OK);	
}

/*---------------------------------------------------------------------------
 * UnlockPortFile -- unlock port file with flock
 *
 *---------------------------------------------------------------------------
 */

UnlockPortFile()
{
	/*if (flock(PortFileFD, LOCK_UN) == ERROR) {
		perror("flock");
		fprintf(stderr, "attempt to unlock port file failed.\n");
		fprintf(stderr, "contact avr.\n");
		return(ERROR);
	} */
 	return(OK);	
}

/*---------------------------------------------------------------------------
 * ReadAddrCache -- read the address cache from port file
 *
 *---------------------------------------------------------------------------
 */

ReadAddrCache()
{
	int i = 0;
	int site, port;

	if (lseek(PortFileFD, 0L, L_SET) < 0) {
		perror("lseek");
		fprintf(stderr, "contact avr.\n");
		exit(ERROR);
	}
	read(PortFileFD, (char *) AddrCache, sizeof(AddrCache));
}

/*---------------------------------------------------------------------------
 * WriteAddrCache -- write address cache out to port file
 *
 *---------------------------------------------------------------------------
 */

WriteAddrCache()
{
	int i;
	char buf[BUFSIZ];
	struct stat stbuf;

	if (lseek(PortFileFD, 0L, L_SET) < 0) {
		perror("lseek");
		fprintf(stderr, "contact avr.\n");
		exit(ERROR);
	}
	write(PortFileFD, (char *) AddrCache, sizeof(AddrCache));
	zero(stbuf);
	fstat(PortFileFD, &stbuf);
	PFLastMtime = stbuf.st_mtime;
}

/*---------------------------------------------------------------------------
 * Send -- send a message to another site
 *
 * Uses the global "sa" structure initialized by Setup().
 *---------------------------------------------------------------------------
 */

Send(buf, SID)
char *buf;
int SID;
{
	int port;
	char msgbuf[MAX_PKT_LEN];

	if ((port = GetPortNumber(SID)) == ERROR) {
		fprintf(stderr, "can't find address of site %d\n", SID);
		return(ERROR);
	}
	sa.sin_port = port;

/*---------------------------------------------------------------------------
:       Update this code to cause the entire message buffer to be transferred
:
:	sprintf(msgbuf, "cs542:%d:%s:", MySID, buf);
:	
:	if (sendto(sMySock, msgbuf, strlen(msgbuf)+1, 0, &sa, sizeof(sa))<0) {
:		perror("sendto");
:		return(ERROR);
:       }
---------------------------------------------------------------------------*/

	sprintf(msgbuf, "cs542:%d", MySID);
	bcopy (buf, &msgbuf[MAX_PKT_LEN-MAX_MSG_LEN], MAX_MSG_LEN);

/* Added cast to 'struct sockaddr *' for argument 5 of sendto() - DM */
	if (sendto(sMySock, msgbuf, MAX_PKT_LEN, 0, (struct sockaddr *)&sa, sizeof(sa))<0) {
		perror("sendto");
		return(ERROR);
	}
}

/*---------------------------------------------------------------------------
 * Receive -- receive a message from another site
 *
 * Blocks until a message is received or a timeout happens.
 *---------------------------------------------------------------------------
 */

Receive(pbuf, pSID)
char *pbuf;
int *pSID;
{

	char msgbuf[MAX_PKT_LEN];
	struct sockaddr_in saFrom;
	int lsa = sizeof(saFrom);
	int len;

	alarm(TIMEOUT_SEC);

/* Added cast to 'struct sockaddr *' for argument 5 of recvfrom() - DM */
	if ((len=recvfrom(sMySock, msgbuf, MAX_PKT_LEN, 0, (struct sockaddr *)&saFrom, &lsa))<0) {
		if (errno == EINTR)
			return(TIMEOUT);
		else {
			perror("recvfrom");
			return(ERROR);
		}
	}
	alarm(0);

/*---------------------------------------------------------------------------
:       Update this code to allow non-character buffer transfer
:
:	if (sscanf(msgbuf, "cs542:%d:%[^:]:", pSID, pbuf) != 2) {
:		fprintf(stderr, "bad message received: '%s'\n", msgbuf);
:		return(ERROR);
:       }
---------------------------------------------------------------------------*/

        bcopy (&msgbuf[MAX_PKT_LEN-MAX_MSG_LEN], pbuf, MAX_MSG_LEN);

	if (sscanf(msgbuf, "cs542:%d:", pSID) != 1) {
		fprintf(stderr, "bad message received: '%s'\n", msgbuf);
		return(ERROR);
	}
	if (*pSID < 0 || *pSID >= MAX_SITES) {
		fprintf(stderr, "message '%s' from bad site %d.\n", 
			msgbuf, *pSID);
		fprintf(stderr, "site IDs must be in range [0,%d].\n",
			MAX_SITES);
		return(ERROR);
	}
	return(OK);
}

/*---------------------------------------------------------------------------
 * GetPortNumber -- get port number of another site
 *
 *---------------------------------------------------------------------------
 */

GetPortNumber(wSID)
int wSID;
{
	static struct stat stbuf;
	int fd;

	zero(stbuf);
	if (LockPortFile() == ERROR)
		return(ERROR);
	if (fstat(PortFileFD, &stbuf) < 0) {
		perror("fstat");
		fprintf(stderr, "fstat on port file failed.\n");
		fprintf(stderr, "contact avr.\n");
		return(ERROR);
	}
	if (UnlockPortFile() == ERROR)
		return(ERROR);
	if (stbuf.st_mtime > PFLastMtime) {
		/* port file has changed */
		PFLastMtime = stbuf.st_mtime;
		if (LockPortFile() == ERROR)
			return(ERROR);
		ReadAddrCache();
		if (UnlockPortFile() == ERROR)
			return(ERROR);
	}
	if (wSID < 0 || wSID >= MAX_SITES) {
		fprintf(stderr, "SID must be in range [0,%d]\n", MAX_SITES);
		return(ERROR);
	}
	return (AddrCache[wSID]);
}

/*---------------------------------------------------------------------------
 * ExitSite -- clean up
 *
 * Must be called when a site goes down to remove its port number
 * from the port file
 *---------------------------------------------------------------------------
 */

ExitSite()
{
	if (LockPortFile() == ERROR)
		return(ERROR);
	ReadAddrCache();
	AddrCache[MySID] = ERROR;
	WriteAddrCache();
	if (UnlockPortFile() == ERROR)
		return(ERROR);
	return(OK);
}
