/*---------------------------------------------------------------------------
:  This file contains utility procedures for site communications
---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------
:  Header files
---------------------------------------------------------------------------*/

#  include  <strings.h>
#  include  "comm.h"
#  include  "simtypes.h"

/*---------------------------------------------------------------------------
:  Procedure to send a message to another site
---------------------------------------------------------------------------*/

send_message (message)

   msg_type *message;
{

/*---------------------------------------------------------------------------
:  Variable definition
---------------------------------------------------------------------------*/

   char *message_translation();
   int  send_status;
   char comment[80] , message_buff[TRANS_MSG_SIZE];

/*---------------------------------------------------------------------------
:  Start of code - log the message and send it out
---------------------------------------------------------------------------*/

   strcpy (message_buff , 
	   message_translation (message->type , message->subtype));
   sprintf (comment , "send %s msg to site %d" , message_buff , message->dest); 
   log_site_info (NO_DUMP , comment);

   if ((send_status = Send (message, message->dest)) == ERROR) {
      if (message->source != SIMULATION_MGR_ID)
	 log_site_info (NO_DUMP , "*** error from Send ***");
      else
	 printf ("*** Unable to send message ***\n");
   }
   return (send_status);
}

/*---------------------------------------------------------------------------
:  Procedure to receive a message from another site
---------------------------------------------------------------------------*/

receive_message (message)

   msg_type *message;
{

/*---------------------------------------------------------------------------
:  Variable definition
---------------------------------------------------------------------------*/

   char *message_translation();
   int rcv_status , msg_source;
   char comment[80] , message_buff[TRANS_MSG_SIZE];

/*---------------------------------------------------------------------------
:  Start of code - loop to receive a message
---------------------------------------------------------------------------*/

   while ((rcv_status = Receive(message , &msg_source)) != OK)
      if (rcv_status != TIMEOUT)  break;

   if (rcv_status == OK) {
      strcpy (message_buff , 
              message_translation (message->type , message->subtype));
      sprintf (comment , "received %s msg from site %d" , message_buff ,
	       message->source); 
      log_site_info (NO_DUMP , comment);
   }
   else
      log_site_info (NO_DUMP , "*** error from Receive ***");

   return (rcv_status);
}

/*---------------------------------------------------------------------------
:  Procedure to translate integer message ID information to string
---------------------------------------------------------------------------*/

char *message_translation  ( type , subtype )

   enum message_types type;
   enum message_subtypes subtype;
{

/*---------------------------------------------------------------------------
:  Variable definition
---------------------------------------------------------------------------*/

   char message_c[TRANS_MSG_SIZE];

/*---------------------------------------------------------------------------
:  Start of code - translate the message type
---------------------------------------------------------------------------*/

   switch (type) {
      case control :
	   strcpy (message_c , "control.");
	   break;
      case xact :
	   strcpy (message_c , "xact.");
	   break;
      case managing :
	   strcpy (message_c , "managing.");
	   break;
      default :
	   strcpy (message_c , "?.");
   }

/*---------------------------------------------------------------------------
:  Translate the message subtype
---------------------------------------------------------------------------*/

   switch (subtype) {
/*********************************/
      case direct_update:
	   strcat (message_c , "direct_update");
	   break ;
      case directory_copy: 
	   strcat (message_c , "directory_copy") ; 
	   break ; 
      case directory_ack:
	   strcat (message_c , "directory_ack") ; 
	   break ; 
/*********************************/
      case recovery_announce :
	   strcat (message_c , "recovery_announce");
	   break;
      case recovery_response :
	   strcat (message_c , "recovery_response");
	   break;
      case recovery_wait :
	   strcat (message_c , "recovery_wait");
	   break;
      case failure_announce :
	   strcat (message_c , "failure_announce");
	   break;
      case clear_fail_locks :
	   strcat (message_c , "clear_fail_locks");
	   break;
      case user :
	   strcat (message_c , "user");
	   break;
      case update :
	   strcat (message_c , "update");
	   break;
      case copier :
	   strcat (message_c , "copier");
	   break;
      case copier_update :
	   strcat (message_c , "copier_update");
	   break;
      case ack :
	   strcat (message_c , "ack");
	   break;
      case commit :
	   strcat (message_c , "commit");
	   break;
      case commit_ack :
	   strcat (message_c , "commit_ack");
	   break;
      case stop :
	   strcat (message_c , "stop");
	   break;
      case revive :
	   strcat (message_c , "revive");
	   break;
      case die :
	   strcat (message_c , "die");
	   break;
      case dump :
	   strcat (message_c , "dump");
	   break;
      case up :
	   strcat (message_c , "up");
	   break;
      case failed :
	   strcat (message_c , "failed");
	   break;
      case allow_recovery :
	   strcat (message_c , "allow_recovery");
	   break;
      case xact_committed :
	   strcat (message_c , "xact_committed");
	   break;
      case xact_aborted :
	   strcat (message_c , "xact_aborted");
	   break;
      case status :
	   strcat (message_c , "status");
	   break;
      default :
	   strcat (message_c , "?");
	   break;
   }
   return (message_c);
}
