/* miscutil.c - contains functions used through out dbsite and mgr routines
		these are provided for the addition of directories to the system
*/

#include <stdio.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/time.h>
#include "comm.h"
#include "simtypes.h"
#include "extdata.h"
#include "direct.h"

/*
 *  directory_mgr - take care of manipulating the directories
 *			check to see if object is at a certain site,
 *			remove an item from a site, add an item to a site.
 */

int directory_mgr (function, site, item)
int	function;
site_id_type	site;
data_item_type	item;
{
int	status = DIRECTORY_ERROR, index;

    if ((index = data_item_mgr (INDEX_ITEM, item, NULL_ITEM_VALUE)) != ITEM_NOT_FOUND) {
	switch (function) {
	    case CREATE_COPY:	/* change directory to show object creation */
		SET_BIT(directories[item].direct,site);
		break;

	    case REMOVE_COPY:	/* change directory to remove object from site*/
		CLEAR_BIT(directories[item].direct,site);
		break;

	    case CHECK_FOR_COPY:
		if (TEST_BIT(directories[item].direct,site))
		    status = COPY_PRESENT;
		else
		    status = COPY_NOT_PRESENT;
		break;

	    default:
		printf ("*** Illegal directory function ***\n");
		break;
	}
    }

    return (status);
}
/*
 *  dump_db -  dump the contents of the data base in a formatted way
 *		item number, value, fail locks, directory  are all printed
 */

dump_db(comment)
{

#  define  MAX_OUT  (80/(10+2*num_sites))
#  define  NUM_TIME_CHARS  24
int	index, i, num_left, num_out;
char	fail_locks_string[MAX_NUM_SITES+1];
char	directory_string[MAX_NUM_SITES+1];
char	*time_ptr, time_string[NUM_TIME_CHARS+1];
/*DM				ctime does not take a struct timeval..
struct	timeval tp;
struct	timezone tzp;
*/
site_id_type	site;

time_t tp = time(0);
/*DM	   gettimeofday(&tp, &tzp); */

	   time_ptr = ctime (&tp);
	   strncpy(time_string, time_ptr, NUM_TIME_CHARS);

	   printf ("%24.24s : %s\n" , time_string , comment);
	   printf ("   actual session number:  %d\n" , actual_session_num);
	   printf ("   session vector (site|state|session num):\n      ");
	   for (site = FIRST_SITE_ID ; site < num_sites ; site++)
	      printf ("%1d|%c|%-2.2d " , site , session_vector[site].state ,
		      session_vector[site].snum);

           printf ("\n   database info (item|value|fail-locks|directory):\n");
	   fail_locks_string[num_sites] = '\0';
	   directory_string[num_sites] = '\0';
	   index = 0;
	   while (index < num_db_items) {
	      num_left = num_db_items - index;
	      num_out =  (num_left > MAX_OUT ? MAX_OUT : num_left);
	      for (i = index ; i < index+num_out ; i++) {
		 for (site = 0 ; site < num_sites ; site++)
	            if (fail_locks_mgr (CHECK_FOR_SITE , site , i) ==
			FAIL_LOCK_SET)
		       sprintf (&fail_locks_string[site] , "%1.1d" , 
				(site<10 ? site : (site % 10)));  
		    else
		       fail_locks_string[site] = '-';

		 for (site = 0 ; site < num_sites ; site++)
	            if (directory_mgr (CHECK_FOR_COPY , site , i) ==
			COPY_PRESENT)
		       sprintf (&directory_string[site] , "%1.1d" , 
				(site<10 ? site : (site % 10)));  
		    else
		       directory_string[site] = '-';

		 printf ("  %2d|%3.3d%|%s|%s" , database[i].item ,
			 database[i].value , fail_locks_string, directory_string);
              }  /* end for */
              index += num_out;
	      printf ("\n");
           }  /* end while */
   fflush (stdout);
   return (0);
}
