/*--------------------------------------------------------------------------
:  This procedure processes a control message type for a database site
---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------
:  Header files
---------------------------------------------------------------------------*/

#  include  <stdio.h>
#  include  "comm.h"
#  include  "simtypes.h"
#  include  "extdata.h"
#  include  "direct.h"

/*DM	moved prototype here to avoid scope warnings */

	site_state_type get_site_state();

/*---------------------------------------------------------------------------
:  Procedure entry point
---------------------------------------------------------------------------*/

process_control_msg ( message )
  
   msg_type *message;
{

/*---------------------------------------------------------------------------
:  Start of code - process according to the message subtype
---------------------------------------------------------------------------*/

   switch (message->subtype) { 
        case 	recovery_announce :
		process_recovery_announce( message );
		break;

	case 	recovery_response :
		process_recovery_response( message );
		break;

	case	recovery_wait	  :
		process_recovery_wait( message );
		break;

	case	failure_announce  :
		process_failure_announce( message );
		break;

	case	clear_fail_locks  :
		process_clear_fail_locks( message );
		break;

	case	status :
		process_status ( message );
		break;
	case    directory_ack: 
		bcopy(message->c.ctl_direct, directories, 
				sizeof(directory_type)) ; 
		break ; 

	default:
		log_site_info(NO_DUMP,"***unknown control msg subtype!***");
		break;
   }

   return (0);
}

/*------------------------------------------------------------------------
 * process_recovery_announce checks the site's session vector to see if
 * 	a multiple site failure occurred by counting the number of sites
 * 	in the recovery_wait state. If all sites that 
 *------------------------------------------------------------------------
 */
   process_recovery_announce( message )
	msg_type	*message;
   {

/*------------------------------------------------------------------------
 * Local variables
 *------------------------------------------------------------------------
 */
	int 	no_sites_up=0;		/* number of sites up that were	*/
					/* up when I went down 		*/
	int	no_sites_waiting=0;	/* number of sites currently in */
				    	/* recovery_wait state		*/	
	int 	highest_site_up=0;	/* highest ranked site currently up */
        site_state_type state;		/* temporary variable 		*/
        site_state_type old_state;	/* temporary variable 		*/
	int	i;			/* index			*/


	state = get_site_state(MySID);
	 if (state == SITE_UP)  
		set_site_state(message->source,SITE_UP);
		/* send_recovery_response(message->source); */
		/* the managing.allow_recovery is doing this */
	 else { 
		if (state == SITE_RECOVERY_WAIT) {
		/* see if the other sites are waiting for me to come up */
			for(i=FIRST_SITE_ID;i<num_sites;i++) {
				if (i != MySID) {
					state = get_site_state(i);
					old_state = get_old_site_state(i);
					if (old_state == SITE_UP) {
					     no_sites_up++;
				       	     highest_site_up = i;
				       	     if (state == SITE_RECOVERY_WAIT)
						no_sites_waiting++;
					     else if (message->source == i) {
						no_sites_waiting++;
						set_site_state(i,
							SITE_RECOVERY_WAIT);
					     }
					} /* end if */
				} /* end if */
			} /* end for */ 

			if ((no_sites_waiting == no_sites_up) &&
			    (MySID >= highest_site_up))  {
				/* mark all waiting sites as up	*/
 printf("I am the highest one recovering\n");
				for(i=FIRST_SITE_ID;i<num_sites;i++) {
					state = get_site_state(i);
					if(state == SITE_RECOVERY_WAIT)  
						set_site_state(i,SITE_UP);
				} 
				/* put self up	*/
				set_site_state(MySID,SITE_UP);
				/* send f_locks and s_vector to ALL up sites */
				for(i=FIRST_SITE_ID;i<num_sites;i++)
				  if ( (get_site_state(i) == SITE_UP ) &&
				      ( i != MySID) )
				    send_recovery_response((site_id_type)i);
			}
			else {
printf("I must wait for other sites\n");
				/* there are sites that I must wait for */
				send_recovery_announce(message->source);
				set_site_state(message->source, 
						SITE_RECOVERY_WAIT);
			}
		}
		else send_managing_failed(message->source);
		/* since I am not up and I am not waiting, I must be down */
	 } 
   }


/*------------------------------------------------------------------------
 * process_recovery_response
 *------------------------------------------------------------------------
 */
   process_recovery_response( message )
	msg_type	*message;
   {
	site_id_type 	i;
	site_state_type	state;
	site_state_type	get_site_state();

	if (get_site_state(MySID) == SITE_RECOVERY_WAIT) {
		/* get session vector and fail locks from message */
		bcopy(message->c.ctl.s_vector, session_vector,
	      		sizeof(session_vector_type));
		bcopy(message->c.ctl.f_locks, fail_locks,
	      		sizeof(fail_locks_type));
		set_site_state(MySID,SITE_UP);
		/* mark all waiting sites as up	*/
		for(i=FIRST_SITE_ID;i<num_sites;i++) {
			state = get_site_state(i);
			if(state == SITE_RECOVERY_WAIT)  
				set_site_state(i,SITE_UP);
		} 
                /* log_site_info(DUMP,"site recovered"); for debugging */
	}
	else log_site_info(NO_DUMP,"***unsolicited recovery.response***");
	/* get the directories too */
	receive_directory() ; 
   }

/*------------------------------------------------------------------------
 * process_recovery_wait
 *------------------------------------------------------------------------
 */
   process_recovery_wait( message )
	msg_type	*message;
   {
	if(get_site_state(MySID) != SITE_RECOVERY_WAIT)
		log_site_info(NO_DUMP,"*** recd unexpected recovery_wait ***");
   }

/*------------------------------------------------------------------------
 * process_failure_announce
 *------------------------------------------------------------------------
 */
   process_failure_announce( message )
	msg_type	*message;
   {
	int	i,k;

	/* update my session vector */
	for(i=FIRST_SITE_ID; i<num_sites; i++)	{
		if ((message->c.ctl.s_vector[i].state == SITE_DOWN) &&
		    (get_site_state(i) != SITE_DOWN))   {
			/* mark it as being down */
			set_site_state(i,SITE_DOWN);
                	/* set fail locks for this site for the last xact */
			if (last_update != NULL) 
			  for (k=0; k<last_update->c.db.num_items; k++) 
				fail_locks_mgr(SET_FOR_SITE,i,
					       last_update->c.db.item[k]);
		}
	}
   }

/*------------------------------------------------------------------------
 * process_clear_fail_locks
 *------------------------------------------------------------------------
 */
   process_clear_fail_locks( message )
	msg_type	*message;
   {
	int	i;

	for(i = 0; i < message->c.db.num_items; i++)
		fail_locks_mgr(CLEAR_FOR_SITE,message->source,
			message->c.db.item[i]);
   }

/*-------------------------------------------------------------------------
 * process_status
 *-------------------------------------------------------------------------
 */
   process_status( message )
	msg_type	*message;
 {
	site_state_type state,get_site_state();

	state = get_site_state(MySID);
	switch (state) 
	{
	case   SITE_DOWN :
			send_managing_failed(message->source);
			break;

	case   SITE_RECOVERY_WAIT :
			send_recovery_announce(message->source);
			break;

	case   SITE_UP :
			send_managing_up(message->source);
			break;

	default : log_site_info(NO_DUMP,"***recd bad control.status mesg ***");
	}
}
