/*---------------------------------------------------------------------------
:  This file contains procedures to process a xact message type for a 
:  database site
---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------
:  Header files
---------------------------------------------------------------------------*/

#  include  <stdio.h>
#  include  "comm.h"
#  include  "simtypes.h"
#  include  "extdata.h"
/*************/
#  include  "direct.h"
/*************/
/*---------------------------------------------------------------------------
:  process_xact_msg initially handles the message and determines next action
---------------------------------------------------------------------------*/

process_xact_msg ( message )
  
   msg_type *message;
{

/*---------------------------------------------------------------------------
:  Local variable definition
---------------------------------------------------------------------------*/

   site_state_type get_site_state();
   msg_type *reply_msg, *fl_msg, *dir_msg;

/*---------------------------------------------------------------------------
:  Start of code - process according to the message subtype if site is up
---------------------------------------------------------------------------*/

   if (get_site_state (MySID) == SITE_UP)

      switch (message->subtype) {
         case  copier :
	       process_copier_xact (message);
	       break;

         case  update :
               process_update_xact (message);
	       break;

         case  user :
	       process_user_xact (message);
	       break;
	 case direct_update: 
	       if (process_direct_update(message, &fl_msg, &dir_msg) == TRUE)
		   get_new_copies(dir_msg, fl_msg) ; 
	       break ; 

         default :
	       log_site_info (NO_DUMP , "*** unsolicited2 message ***");
	       break;
      }  /* end switch */

/*---------------------------------------------------------------------------
:  Site down so no processing done.  Inform another db site if necessary
---------------------------------------------------------------------------*/

   else {
      log_site_info (NO_DUMP , "this site down => no processing");
      if (message->source != SIMULATION_MGR_ID) {
	 reply_msg = (msg_type *) malloc (sizeof (msg_type));
         reply_msg->source = MySID;
	 reply_msg->dest = message->source;
	 reply_msg->type = managing;
	 reply_msg->subtype = failed;
	 send_message (reply_msg);
	 free (reply_msg);
      }
   }
   return (0);
}

/*---------------------------------------------------------------------------
:  process_copier_xact sends the requested data-item values to the requestor.
---------------------------------------------------------------------------*/

process_copier_xact ( msg )
 
   msg_type *msg;
{

/*---------------------------------------------------------------------------
:  Local variable definition
---------------------------------------------------------------------------*/

   int send_message();
   boolean dump_needed = FALSE;
   int index;
   msg_type *reply_msg;
   data_item_type cur_item_value;
   char comment[40];

/*---------------------------------------------------------------------------
:  Start of code - set the message fields and look up requested data-items
---------------------------------------------------------------------------*/

   reply_msg = (msg_type *) malloc (sizeof(msg_type));
   reply_msg->source = MySID;
   reply_msg->dest = msg->source;
   reply_msg->type = xact;
   reply_msg->subtype = copier_update;
   reply_msg->c.db.num_items = 0;

   for (index = 0 ; index < msg->c.db.num_items ; index++)
      if (fail_locks_mgr (CHECK_FOR_SITE , MySID , msg->c.db.item[index]) ==
          FAIL_LOCK_CLEAR) {
	 reply_msg->c.db.num_items ++;
	 reply_msg->c.db.item[index] = msg->c.db.item[index];
         data_item_mgr (READ_ITEM , msg->c.db.item[index] , &cur_item_value);
	 reply_msg->c.db.value[index] = cur_item_value;
      }
      else {
         dump_needed = TRUE;
	 sprintf (comment , "*** Copier error for %d ***" , 
		  msg->c.db.item[index]);
         log_site_info (NO_DUMP , comment);
      }

/*---------------------------------------------------------------------------
:  Dump info if any error occurred and send reply to requestor  
---------------------------------------------------------------------------*/

   if (dump_needed)  log_site_info (DUMP , "dump after error");
   send_message (reply_msg);
   free (reply_msg);
   return (0);
   }

/*---------------------------------------------------------------------------
:  process_update_xact processes the update message which was sent to this
:  site from another site as the result of a user transaction
---------------------------------------------------------------------------*/

process_update_xact ( upd_msg )
 
   msg_type *upd_msg;
{

/*---------------------------------------------------------------------------
:  Local variable definition
---------------------------------------------------------------------------*/

   int send_message() , receive_message() , announce_site_failure();
   site_state_type set_site_state() , get_site_state();
   msg_type *cur_msg;
   boolean need_final_reply = TRUE;
   int index;

/*---------------------------------------------------------------------------
:  Start of code - acknowledge reception of the update message
---------------------------------------------------------------------------*/

   cur_msg = (msg_type *) malloc (sizeof(msg_type));
   cur_msg->source = MySID;
   cur_msg->dest = upd_msg->source;
   cur_msg->type = xact;
   cur_msg->subtype = ack;
   send_message (cur_msg);

/*---------------------------------------------------------------------------
:  Get a reply from the updater (i.e., the site propagating the update)
---------------------------------------------------------------------------*/

   receive_message (cur_msg);

   switch (cur_msg->type) {

/*---------------------------------------------------------------------------
:     First possibility:  get xact.commit from updater after a successful
:     first round of update
---------------------------------------------------------------------------*/

      case xact :
	   if (cur_msg->subtype != commit) {
	      need_final_reply = FALSE;
	      log_site_info (NO_DUMP , "*** unexpected message ***");
	   }
           else {

/*---------------------------------------------------------------------------
:             Commit the transaction if a failure is not scheduled
---------------------------------------------------------------------------*/

	      if (failure_schedule != FAIL_AFTER_UPD_RECV) {
		 commit_db_values (upd_msg);
		 cur_msg->type = xact;
		 cur_msg->subtype = commit_ack;
              }

/*---------------------------------------------------------------------------
:             This site has failed after the first round of an update
---------------------------------------------------------------------------*/

              else {
                 set_site_state (MySID , SITE_DOWN);
                 log_site_info (DUMP , "failing after update receive");
		 cur_msg->type = managing;
		 cur_msg->subtype = failed;
	      }
	   }
           break;

/*---------------------------------------------------------------------------
:     Second possibility:  get control.failure_announce from updater if it
:     detected that another site failed during the first round of update.
:     The transaction in progress is discarded.
---------------------------------------------------------------------------*/

      case control :
	   need_final_reply = FALSE;
	   if (cur_msg->subtype == failure_announce)
	      process_failure_announce (cur_msg);
           else
	      log_site_info (NO_DUMP , "*** unexpected message ***");
	   break;

/*---------------------------------------------------------------------------
:     Third possibility:  get managing.failed from updater if it failed   
:     after the first round of update.  The update is discarded.
---------------------------------------------------------------------------*/

      case managing :
	   need_final_reply = FALSE;
	   if (cur_msg->subtype != failed)
	      log_site_info (NO_DUMP , "*** unexpected message ***");
           else if (failure_schedule != FAIL_AFTER_UPD_RECV) {
              set_site_state (upd_msg->source , SITE_DOWN);
	      announce_site_failure ();
	   }
           else  /* this site has a scheduled failure */ {
              set_site_state (MySID , SITE_DOWN);
              log_site_info (DUMP , "failing after update receive");
	   }
	   break;

      default :
	   need_final_reply = FALSE;
	   log_site_info (NO_DUMP , "*** unexpected message ***");
	   break;
   }  /* end switch */

   if (need_final_reply) {
      cur_msg->source = MySID;
      cur_msg->dest = upd_msg->source;
      send_message (cur_msg);
   }
   return (0);
}

/*---------------------------------------------------------------------------
:  process_user_xact processes a user transaction for the database
---------------------------------------------------------------------------*/

process_user_xact ( user_msg )
 
   msg_type *user_msg;
{

/*---------------------------------------------------------------------------
:  Local variable definition
---------------------------------------------------------------------------*/

   int insure_current_database() , send_message() , receive_message() ,
       announce_site_failure() , data_item_mgr() , fail_locks_mgr();
   site_state_type set_site_state() , get_site_state();
   site_id_type site;
   msg_type *upd_msg , *snd_msg , *rcv_msg , *abort_msg;
   data_value_type no_value;
   int write_set , i, index;
   boolean commit_transaction = TRUE , failure_detected;
   char comment[30];
   boolean item_exists; 

/*---------------------------------------------------------------------------
:  Start of code - initialize messages and determine if the database 
:  is up to date for all items of xact read set
---------------------------------------------------------------------------*/

   log_message (user_msg);
   abort_msg = (msg_type *) malloc (sizeof (msg_type));
   upd_msg = (msg_type *) malloc (sizeof(msg_type));
   upd_msg->source = MySID;
   upd_msg->type = xact;
   upd_msg->subtype = update;
   upd_msg->c.db.num_items = 0;

   if (insure_current_database (user_msg , abort_msg ) == 
       DB_UP_TO_DATE) {

/*---------------------------------------------------------------------------
:     Put all items and values of the xact write set into the update message
---------------------------------------------------------------------------*/
      write_set = 0;
      for (index = 0 ; index < user_msg->c.ux.num_actions ; index++)
	 if (user_msg->c.ux.op[index].action == WRITE_ITEM)
	    if (data_item_mgr
		   (INDEX_ITEM , user_msg->c.ux.op[index].item , &no_value) !=
		ITEM_NOT_FOUND) {

	       /* check if item exists on an up site */
	       item_exists = FALSE ; 
	       for (i = 0 ; i < num_sites ; i++) { 
		   if (get_site_state(i) == SITE_UP && directory_mgr(
			CHECK_FOR_COPY, i, user_msg->c.ux.op[index].item ) ==
			COPY_PRESENT ) { 
		      item_exists = TRUE ; 
		      break ; 
		   } 
	       } 
	       if (item_exists) { 
	          upd_msg->c.db.item[write_set] = user_msg->c.ux.op[index].item;
                  upd_msg->c.db.value[write_set] = 
				user_msg->c.ux.op[index].value;
	          write_set ++;
	       } 
	       else { 
		  write_set = 0 ; 
		  commit_transaction = FALSE ; 
		  printf("All copies of item%d are down\n", 
				user_msg->c.ux.op[index].item ) ; 
		  break ; 
	       } 
	    }
	    else {
	       sprintf (comment , "*** bad item in write set:  %-d ***" ,
			user_msg->c.ux.op[index].item);
	       log_site_info (NO_DUMP , comment);
	    }

/*---------------------------------------------------------------------------
:     If write set is not empty, then send xact.update to each UP site and
:     check the other site's reply (it should be xact.ack or managing.failed)
---------------------------------------------------------------------------*/

      if (write_set > 0) {
         upd_msg->c.db.num_items = write_set;
	 rcv_msg = (msg_type *) malloc (sizeof(msg_type));

         for (site = FIRST_SITE_ID ; site < num_sites ; site++) 
	    if ((site != MySID) && (get_site_state (site) == SITE_UP)) {
               upd_msg->dest = site;
	       if (send_message (upd_msg) != ERROR) {
		  receive_message (rcv_msg);

                  if ((rcv_msg->type == managing) &&
		      (rcv_msg->subtype == failed)) {
                     commit_transaction = FALSE;
		     set_site_state (rcv_msg->source , SITE_DOWN);
		     announce_site_failure ();
		     check_direct(); 
		     break;
		  }
		  else if ((rcv_msg->type != xact) || (rcv_msg->subtype != ack))
		     log_site_info (NO_DUMP , "*** unexpected message ***"); 
               }
	    }

/*---------------------------------------------------------------------------
:        Round one completed - if xact has not been aborted then either
:        commit at other sites or go into scheduled failure.  If committing,
:        other sites should respond with xact.commit_ack or managing.failed.
---------------------------------------------------------------------------*/

         if (commit_transaction) {
	    snd_msg = (msg_type *) malloc (sizeof(msg_type));
	    snd_msg->source = MySID;
	    if (failure_schedule != FAIL_AFTER_UPD_SEND) {
	       snd_msg->type = xact;
	       snd_msg->subtype = commit;
	    }
            else {
	       commit_transaction = FALSE;
	       set_site_state (MySID , SITE_DOWN);
	       log_site_info (DUMP , "failing after update send");
	       snd_msg->type = managing;
	       snd_msg->subtype = failed;
	    }
          
	    failure_detected = FALSE;
            for (site = FIRST_SITE_ID ; site < num_sites ; site++) 
	       if ((site != MySID) && (get_site_state (site) == SITE_UP)) {
		  snd_msg->dest = site;
		  if ((send_message (snd_msg) != ERROR) &&
		      (snd_msg->subtype == commit)) {

		     receive_message (rcv_msg);
		     if ((rcv_msg->type == managing) &&
			 (rcv_msg->subtype == failed)) {
			failure_detected = TRUE;
			set_site_state (rcv_msg->source , SITE_DOWN);
		     }
                     else if ((rcv_msg->type != xact) ||
			      (rcv_msg->subtype != commit_ack))
			log_site_info (NO_DUMP , "*** unexpected message ***");
                  }
               }

            free (snd_msg);
	    free (rcv_msg);
            if (failure_detected)  { 
	       announce_site_failure ();
	       check_direct() ; 
	    } 
	 }  /* end if */
      }  /* end if */
   }  /* end if */

/*---------------------------------------------------------------------------
:  Fail locks on read set items could not be cleared so abort
---------------------------------------------------------------------------*/

   else {
      commit_transaction = FALSE;
      log_site_info (NO_DUMP , "db obsolete => transaction aborted");
   }

/*---------------------------------------------------------------------------
:  Update this site's database if necessary and then inform the simulation
:  manager on the outcome of the transaction
---------------------------------------------------------------------------*/

   if (commit_transaction) {
      commit_db_values (upd_msg);
      upd_msg->dest = user_msg->source ; /* SIMULATION_MGR_ID;*/
      upd_msg->type = managing;
      upd_msg->subtype = xact_committed;
      upd_msg->c.db.num_copier_xacts = num_copier;
      upd_msg->c.db.num_pseudo_xacts = num_pseudo;
      num_copier = num_pseudo = 0;
      send_message (upd_msg);
   }
   else {
      abort_msg->dest = user_msg->source ; /* SIMULATION_MGR_ID;*/ 
      abort_msg->type = managing;
      abort_msg->subtype = xact_aborted;
      abort_msg->c.db.num_copier_xacts = num_copier;
      abort_msg->c.db.num_pseudo_xacts = num_pseudo;
      num_copier = num_pseudo = 0;
      send_message (abort_msg);
   }
   free (upd_msg);
   free (abort_msg);
   return (0);
}

/*---------------------------------------------------------------------------
:  insure_current_database verifies that the database of a site has 
:  data-items which are current for the user xact to be processed
:  user transaction and tries to update such items by issuing copier xacts.
---------------------------------------------------------------------------*/

insure_current_database (user_msg , clear_msg )

   msg_type *user_msg , *clear_msg;
{

/*---------------------------------------------------------------------------
:  Local variable definition
---------------------------------------------------------------------------*/

   int send_message() , receive_message() , log_site_info() ,
       announce_site_failure() , data_item_mgr() , fail_locks_mgr();
   int num_locked , index , num_requested;
   data_value_type cur_item_value;
   boolean db_check_complete , failure_detected;
   site_id_type site;
   msg_type *rcv_msg , *copy_msg, *fl_msg, *dir_msg;
   char comment[40];

   typedef struct {
	   site_id_type    site;
	   data_item_type  item;
	   } locked_type;

   locked_type locked[MAX_NUM_ITEMS];
/***********/
   typedef struct {
	   site_id_type    site;
	   data_item_type  item;
	   boolean	   found ; 
	   } missing_type;
   missing_type missing[MAX_NUM_ITEMS];  
   int 	missing_requested, num_missing, num_not_present, i ; 
   msg_type *miss_msg ; 
   boolean found[MAX_NUM_XACT_ACTIONS] ; 
/***********/


/*---------------------------------------------------------------------------
:  Start of code - initialize a message for clearing fail-locks and
:  loop to verify the database
---------------------------------------------------------------------------*/

   clear_msg->source = MySID;
   clear_msg->type = control;
   clear_msg->subtype = clear_fail_locks;
   clear_msg->c.db.num_items = 0;

   for (index = 0 ; index < MAX_NUM_XACT_ACTIONS ; index++) 
       found[index] = FALSE ; 

   db_check_complete = FALSE;
   while (! db_check_complete) {

/*---------------------------------------------------------------------------
:     Build a list of data-items which are fail locked and need to be copied 
---------------------------------------------------------------------------*/
      num_locked = 0;
      num_missing = 0 ; 
      num_not_present = 0 ; 
      for (index = 0 ; index < user_msg->c.ux.num_actions ; index++)
	 if (user_msg->c.ux.op[index].action == READ_ITEM) { 
	     if (directory_mgr(CHECK_FOR_COPY, MySID, 
			user_msg->c.ux.op[index].item) == COPY_PRESENT ) { 
	        if (fail_locks_mgr (CHECK_FOR_SITE , MySID ,
			    user_msg->c.ux.op[index].item) == FAIL_LOCK_SET) 
	    	   locked[num_locked++].item = user_msg->c.ux.op[index].item;
	     } 
	     else if (found[index] == FALSE ) 
		 missing[num_missing++].item = user_msg->c.ux.op[index].item;
	 } 

/*---------------------------------------------------------------------------
:     Attempt to match each locked item with a site which has a good copy.
:     If all sites have an item locked then no more checking is necessary. 
---------------------------------------------------------------------------*/

      for (index = 0 ; index < num_locked ; index++) {
	 locked[index].site = NULL_SITE_ID;
	 for (site = 0 ; site < num_sites ; site++)
            if ((get_site_state (site) == SITE_UP) &&
	        (fail_locks_mgr (CHECK_FOR_SITE , site , locked[index].item) ==
                 FAIL_LOCK_CLEAR) && directory_mgr(CHECK_FOR_COPY, site, 
				locked[index].item) == COPY_PRESENT) {

               locked[index].site = site;
	       break;   /* site found for this item => break for loop */
	    }
         if (locked[index].site == NULL_SITE_ID) {
	    sprintf (comment , "all sites locked for %d" , locked[index].item);
	    log_site_info (NO_DUMP , comment);
	    db_check_complete = TRUE;
	    break;   /* current item locked everywhere => skip other items */
         }
      }  /* end for */

/*---------------------------------------------------------------------------
:     Get good copies of items from the other sites if necessary
---------------------------------------------------------------------------*/

      if ((num_locked > 0) && (! db_check_complete)) {
	 failure_detected = FALSE;
         num_requested = 0;
	 rcv_msg = (msg_type *) malloc (sizeof(msg_type));
	 copy_msg = (msg_type *) malloc (sizeof(msg_type));
	 copy_msg->source = MySID;
	 copy_msg->type = xact;
	 copy_msg->subtype = copier;

	 for (site = FIRST_SITE_ID ; site < num_sites ; site++) {
	    if (site == MySID) continue;
	    copy_msg->c.db.num_items = 0; 

/*---------------------------------------------------------------------------
:           Make a list of all items to be requested from the current site
---------------------------------------------------------------------------*/

            for (index = 0 ; index < num_locked ; index++)
	       if (site == locked[index].site) {
		  copy_msg->c.db.item[copy_msg->c.db.num_items] =
		     locked[index].item;
		  copy_msg->c.db.num_items ++;
	       }

/*---------------------------------------------------------------------------
:           If the site has some good copies then issue the copier xact
:           and process the reply
---------------------------------------------------------------------------*/

            if (copy_msg->c.db.num_items > 0) {
               num_requested += copy_msg->c.db.num_items;
	       copy_msg->dest = site;
	       send_message (copy_msg);
		  
               receive_message (rcv_msg);
	       
/*---------------------------------------------------------------------------
:              Site responded with items so update the database and add
:              items to the list of cleared items
---------------------------------------------------------------------------*/

	       if ((rcv_msg->type == xact) && 
		   (rcv_msg->subtype == copier_update)) {

		  num_copier ++;
                  for (index = 0 ; index < rcv_msg->c.db.num_items ; index++) {
		     cur_item_value = rcv_msg->c.db.value[index];
		     data_item_mgr (WRITE_ITEM , rcv_msg->c.db.item[index] ,
				    &cur_item_value);
                     fail_locks_mgr (CLEAR_FOR_SITE , MySID ,
				     rcv_msg->c.db.item[index]);
                     clear_msg->c.db.item[clear_msg->c.db.num_items] =
			rcv_msg->c.db.item[index];
                     clear_msg->c.db.num_items ++;
		  }  /* end for */
               } 
               
/*---------------------------------------------------------------------------
:              Other site has failed so update its state 
---------------------------------------------------------------------------*/

               else if ((rcv_msg->type == managing) &&
			(rcv_msg->subtype == failed)) {
		  failure_detected = TRUE;
                  set_site_state (rcv_msg->source , SITE_DOWN);
	       }
               else
		  log_site_info (NO_DUMP , "*** unexpected message ***");
            }  /* end if

/*---------------------------------------------------------------------------
:           If a request has been made for every locked item then skip the
:           other sites
---------------------------------------------------------------------------*/

            if (num_locked == num_requested)  break;
	 }  /* end for */

/*---------------------------------------------------------------------------
:        Current set of locked items processed.
---------------------------------------------------------------------------*/

         free (rcv_msg);
         free (copy_msg);
      } /* end if */

/*---------------------------------------------------------------------------
:     Attempt to match each missing item with a site which has a good copy.
:     If all sites have an item missing then no more checking is necessary. 
---------------------------------------------------------------------------*/

      for (index = 0 ; index < num_missing ; index++) {
	 missing[index].site = NULL_SITE_ID;
	 for (site = 0 ; site < num_sites ; site++)
            if ((get_site_state (site) == SITE_UP) &&
	        (directory_mgr (CHECK_FOR_COPY , site , missing[index].item) ==
                 COPY_PRESENT)) {

               missing[index].site = site;
	       break;   /* site found for this item => break for loop */
	    }
         if (missing[index].site == NULL_SITE_ID) {
	    sprintf (comment , "all sites missing for %d" ,missing[index].item);
	    log_site_info (NO_DUMP , comment);
	    db_check_complete = TRUE;
	    break;   /* current item locked everywhere => skip other items */
         }
      }  /* end for */

/*---------------------------------------------------------------------------
:     Get existing copies of items from the other sites if necessary
---------------------------------------------------------------------------*/

      if ((num_missing > 0) && (! db_check_complete)) {
	 failure_detected = FALSE;
         missing_requested = 0;
	 rcv_msg = (msg_type *) malloc (sizeof(msg_type));
	 miss_msg = (msg_type *) malloc (sizeof(msg_type));
	 miss_msg->source = MySID;
	 miss_msg->type = xact;
	 miss_msg->subtype = user;

	 for (site = FIRST_SITE_ID ; site < num_sites ; site++) {
	    if (site == MySID) continue;
	    miss_msg->c.ux.num_actions = 0; 

/*---------------------------------------------------------------------------
:           Make a list of all items to be requested from the current site
---------------------------------------------------------------------------*/

            for (index = 0 ; index < num_missing ; index++){ 
	       if (site == missing[index].site) { 
		  miss_msg->c.ux.op[miss_msg->c.ux.num_actions].action = 
						READ_ITEM ; 
		  miss_msg->c.ux.op[miss_msg->c.ux.num_actions++].item =
		     missing[index].item;
	       } 
	    } 

/*---------------------------------------------------------------------------
:           If the site has some good copies then issue the copier xact
:           and process the reply
---------------------------------------------------------------------------*/

            if (miss_msg->c.ux.num_actions > 0) {
               missing_requested += miss_msg->c.ux.num_actions;
	       miss_msg->dest = site;
	       send_message (miss_msg);
		  
               receive_message (rcv_msg);
	       
/*---------------------------------------------------------------------------
:             keep receiving messages as long as they are failure_announce 
: 	      or clear_fail_locks.  
---------------------------------------------------------------------------*/
	       while ((rcv_msg->type == control && rcv_msg->subtype == 
			failure_announce) || (rcv_msg->type == control && 
			rcv_msg->subtype == clear_fail_locks) || 
			(rcv_msg->type == xact && rcv_msg->subtype == copier)||
			(rcv_msg->type == xact && rcv_msg->subtype ==
					direct_update)) { 
	       	      if (rcv_msg->type == control && rcv_msg->subtype == 
				failure_announce)  
			   process_failure_announce(rcv_msg) ; 
		      else if (rcv_msg->type == xact && rcv_msg->subtype
					== copier ) 
				 process_copier_xact(rcv_msg) ; 
		           else if (rcv_msg->type == xact && rcv_msg->
					   subtype == direct_update) { 
				   if (process_direct_update(rcv_msg,
						&fl_msg, &dir_msg)==TRUE)
					   get_new_copies(dir_msg, fl_msg); 
				   } 
		      receive_message(rcv_msg) ; 
	       } 
/*---------------------------------------------------------------------------
:              Site responded with another message.  
---------------------------------------------------------------------------*/

	       if ((rcv_msg->type == managing) && 
		   (rcv_msg->subtype == xact_aborted)) {
	  /* only way it can receive this is when the database is obsolete */
		   num_pseudo += rcv_msg->c.db.num_pseudo_xacts + 1;
		   num_copier += rcv_msg->c.db.num_copier_xacts;
		   db_check_complete = TRUE ; 
		   break ; 
	       } 
/*---------------------------------------------------------------------------
:              Other site has failed so update its state 
---------------------------------------------------------------------------*/
               else if ((rcv_msg->type == managing) &&
			(rcv_msg->subtype == failed)) {
		       failure_detected = TRUE;
                       set_site_state (rcv_msg->source , SITE_DOWN);
	            }
                    else if (rcv_msg->type == managing && rcv_msg->subtype ==
				xact_committed ) { 
		   	    num_pseudo += rcv_msg->c.db.num_pseudo_xacts + 1;
		   	    num_copier += rcv_msg->c.db.num_copier_xacts;
			    for (index = 0;index < miss_msg->c.ux.num_actions ; 
							index++) { 
				for (i = 0 ; i < user_msg->c.ux.num_actions ; 
								i++) { 
				    if (user_msg->c.ux.op[i].item == 
					 miss_msg->c.ux.op[index].item ) 
				       found[i] = TRUE ; 
				} 
			    } 
			 } 
			 else 
	                        log_site_info (NO_DUMP , 
				            "*** unexpected message ***");
            }  /* end if */
  
/*---------------------------------------------------------------------------
:           If a request has been made for every locked item then skip the
:           other sites
---------------------------------------------------------------------------*/

            if (missing_requested  == num_missing)  break;
	 }  /* end for */

/*---------------------------------------------------------------------------
:        Current set of locked items processed.
---------------------------------------------------------------------------*/

         free (rcv_msg);
	 free (miss_msg) ; 
         if (failure_detected)  { 
	    announce_site_failure ();
	    check_direct() ; 
	 } 
      }  /* end if */

      if (num_locked == 0 && num_missing == 0 )  
	  db_check_complete = TRUE;
   }  /* end while */

/*---------------------------------------------------------------------------
:  Clear fail locks on other sites if necessary
---------------------------------------------------------------------------*/

   if (clear_msg->c.db.num_items > 0) {
      for (site = 0 ; site < num_sites ; site++)
	 if ((site != MySID) && (get_site_state (site) == SITE_UP)) {
	    clear_msg->dest = site;
	    send_message (clear_msg);
	 }
      clear_msg->dest = SIMULATION_MGR_ID ; 
      send_message(clear_msg) ; 
   }
   return ((num_locked == 0) && (num_missing == 0) ? DB_UP_TO_DATE : 
						     DB_OBSOLETE);
}
